/* Simulate the two-dimensional forest-fire model */
 
/* Written December 1995 - August 1996 by Andreas Honecker */
/* Error corrected 25.04.2000 */
/* E-Mail: a.honecker@tu-bs.de */
/* This file may be freely distributed and changed, */
/* provided that your changes are marked clearly */
/* and this and the previous lines remain unchanged */

/* This source contains some inactive remnants of obsolete code */
/* which I used e.g. for debugging purposes. I have not deleted */
/* these parts in comments, because one never knows who might */
/* find them useful. */
 
#include <stdlib.h>
#include <stdio.h>
#include "xdisplay.h"

/* #define random lrand48 */

/* Values of truth */

#ifndef TRUE
  #define TRUE -1
#endif
#ifndef FALSE
  #define FALSE 0
#endif

#define rand() random()
#define srand(seed) srandom(seed)

/* Define the following variable in order to produce an
   X11 output of the state of the system */

static int xdisplay = FALSE;

/* Define the following variable to also determine cluster
   quantities */
/* #define GETCON */

/* Global exit flag */

static int exit_flag = FALSE;

/* Value of f (1 is scaled to 1000000) */

long f_val = 5000;

/* Variable to store the field */

INT *field;

/* Variable to store help copies of field */

#ifdef GETCON
INT *field1;
#endif

/* Size of field */
/* (the width must be set to a multiple of 32) */

static int width=16384;
static int height=16384;
static int word_width;

/* Iterations to equilibrate */

static int equilib = 3000;

/* Total number of iterations */

#define MAX_ITERATIONS 30000

static int iterations = 15000;

/* Values of concentration */

long conc[MAX_ITERATIONS];

/* Local densities */

long long local_dens[257];

/* Corresponding flags */

#define LOC16X16 0xa51234a5
#define LOCNSQ 0x5a43215a

static int local_rows = 1;
static int local_columns = 8;

/* Maximal distance in correlation function */

#define MAXCF 3072

static int maxcf = MAXCF;
static int maxcf_tp = MAXCF;

/* Some magic numbers for testing integrity of dump file */

#ifdef GETCON
 #define MAGICD 0xa5a54321
#else
 #define MAGICD 0x1234a5a5
#endif

/* Stack for bits that are touched durig filling */

struct stack_entry {int row;
                    int column_word;
                    INT mask;
                   };

#define STACKSIZE 1000000

static int stack_cnt;

#ifdef GETCON
struct stack_entry stack[STACKSIZE];
#endif

/* A few auxiliary things to treat large clusters */

struct vert_entry {int y;
                   struct vert_entry *next;
                  };

struct vert_entry **large_cluster;

/* Pseudo memory management */

struct vert_entry vert_space[STACKSIZE];
int vert_space_cnt = 0;

/* This is code for counting bits in a word */

static BYTE bitcnt[65536];

#define bits_in(var) ((bitcnt[(var) & 0xffff] + bitcnt[((var) >> 16) & 0xffff]))

/* Initialize Bit-Count table */

void init_bitcnt()
 {
  int i, j;

  for(i=0; i<65536; i++)
   {
    bitcnt[i] = 0;
    for(j=0; j<16; j++)
     if(i & (1 << j))
         bitcnt[i]++;
   }
 }

/* Initialize field with mean density rho = 1/2 */

void init_field()
 {
  register int i;

  for(i=0; i < (width*height)/32; i++)
     field[i] = rand() + (rand() << 15) + (rand() << 30);
 }

/* Count trees in configuration */

long count_trees()
 {
  int i;
  long count;

  count = 0;
  for(i=0; i < (width*height)/32; i++)
    count += bits_in(field[i]);
  return(count);
 }

/* Clear local density storage */

void clear_local_dens()
 {
  int i;

  for(i=0; i<=256; i++)
    local_dens[i] = 0;
 }

/* Extract local density fluctuations */
/* (Using 16x16 plaquettes) */

void get_local_dens16x16()
 {
  int x, y;
  int yo;
  int counta, countb;
  long index;
  INT tmp_val;

  for(y=0; y<height; y+=16)
   for(x=0; x<word_width; x++)
    {
     counta = countb = 0;
     for(yo=0; yo<16; yo++)
      {
       index = (word_width*(y+yo))+x;
       tmp_val = field[index];
       counta += bitcnt[tmp_val & 0xffff];
       countb += bitcnt[(tmp_val >> 16) & 0xffff];
      }
     local_dens[counta]++;
     local_dens[countb]++;
    }
 }

/* Extract local density fluctuations */
/* (Using plaquettes of size 8*local_clomuns x local_rows) */

void get_local_dens_cr()
 {
  int x, y;
  int yo, xo;
  int count;
  long index;
  INT tmp_val;

  for(y=0; y<height; y+=local_rows)
   for(x=0; x<word_width; x+=local_columns)
    {
     count = 0;
     for(yo=0; yo<local_rows; yo++)
      {
       index = (word_width*(y+yo))+x;
       for(xo = 0; xo<local_columns; xo++)
        {
         tmp_val = field[index+xo];
         count += bitcnt[tmp_val & 0xffff];
         count += bitcnt[(tmp_val >> 16) & 0xffff];
        }
      }
     local_dens[count]++;
    }
 }


/* Filling iteration */

void fill(INT *ptr, int row, int column_word, INT mask)
 {
  int next_row;
  int prev_row;
  INT *next_ptr;
  INT *prev_ptr;

  if(! (*ptr & mask))
     return;

  next_row = row+1;
  if(next_row >= height)   /* Periodic boundary conditions */
    next_row = 0;
  next_ptr = field + word_width*next_row + column_word;

  prev_row = row-1;
  if(prev_row <0)          /* Periodic boundary conditions */
/* ****** Wrong !!!!!
    next_row = height-1;
   ****** Corrected 25.04.2000 */
    prev_row = height-1;
  prev_ptr = field + word_width*prev_row + column_word;

  *ptr ^= mask;
  fill(next_ptr, next_row, column_word, mask);
  fill(prev_ptr, prev_row, column_word, mask);
  if(mask != 1)
   fill(ptr, row, column_word, mask>>1);
  else
   {
    if(column_word < word_width-1)   /* Periodic boundary conditions */
      fill(ptr+1, row, column_word+1, 1<<31);
    else
      fill(ptr+1-word_width, row, 0, 1<<31);
   }
  if(mask != (1<<31))
   fill(ptr, row, column_word, mask<<1);
  else
   {
    if(column_word > 0 )             /* Periodic boundary conditions */
      fill(ptr-1, row, column_word-1, 1);
    else
      fill(ptr-1+word_width, row, word_width-1, 1);
   }
 }

/* Filling iteration with storage of filled pixels */

#ifdef GETCON
void fill_and_save(INT *ptr, int row, int column_word, INT mask)
 {
  int next_row;
  int prev_row;
  INT *next_ptr;
  INT *prev_ptr;

  if(! (*ptr & mask))
     return;

  stack[stack_cnt].row = row;
  stack[stack_cnt].column_word = column_word;
  stack[stack_cnt].mask = mask;

  if(stack_cnt < STACKSIZE)
    stack_cnt++;
  else
    printf("### Error: Fill stack too small ! ###\n");

  next_row = row+1;
  if(next_row >= height)   /* Periodic boundary conditions */
    next_row = 0;
  next_ptr = field + word_width*next_row + column_word;

  prev_row = row-1;
  if(prev_row <0)          /* Periodic boundary conditions */
/* ****** Wrong !!!!! 
    next_row = height-1;
   ****** Corrected 25.04.2000 */
    prev_row = height-1;
  prev_ptr = field + word_width*prev_row + column_word;

  *ptr ^= mask;
  fill_and_save(next_ptr, next_row, column_word, mask);
  fill_and_save(prev_ptr, prev_row, column_word, mask);
  if(mask != 1)
   fill_and_save(ptr, row, column_word, mask>>1);
  else
   {
    if(column_word < word_width-1)   /* Periodic boundary conditions */
      fill_and_save(ptr+1, row, column_word+1, 1<<31);
    else
      fill_and_save(ptr+1-word_width, row, 0, 1<<31);
   }
  if(mask != (1<<31))
   fill_and_save(ptr, row, column_word, mask<<1);
  else
   {
    if(column_word > 0 )             /* Periodic boundary conditions */
      fill_and_save(ptr-1, row, column_word-1, 1);
    else
      fill_and_save(ptr-1+word_width, row, word_width-1, 1);
   }
 }
#endif

/* Iteration step */

void iteration_step()
 {
  int row;
  int column_word;
  int column_bit;
  INT mask;
  INT *ptr;

  row = rand() % height;
  column_word = rand() % word_width;
  column_bit = rand() & 31;
  mask = 1 << column_bit;
  ptr = field + word_width*row + column_word;
  if(*ptr & mask)
   {
/*
    printf("Place is occupied\n");
*/
    if((random() % 1000000) < f_val)   /* Destroy cluster with probability f  = f_val/1000000 */
     {
      fill(ptr, row, column_word, mask);
      if(xdisplay)
        exit_flag |= updateXdisplay();
     }
   }
  else
   {
/*
    printf("Place is empty\n");
*/
    *ptr ^= mask;		/* Grow tree with probability one */
   }
 }

/* Get correlation function */
 
void get_cf(INT *state, long long *cf)
 {
   int row, column;
   INT *ptr;
   register int row_ofs;
   register INT v;

putchar('*');
fflush(stdout);

   for(column=0; column < word_width; column++)
    for(row=0; row < height; row++)
     {
      ptr = field + column + row*word_width;
      v = *ptr;
      if(v)
       {
        for(row_ofs = 0; row_ofs <= maxcf_tp; row_ofs ++)
         {
          if(row + row_ofs < height)
           cf[row_ofs] += bits_in(v & *(ptr + row_ofs*word_width));
          else
           cf[row_ofs] += bits_in(v & *(ptr - (height - row_ofs)*word_width));
         }
       }
     }
 }

/* Get connected correlation function */

#ifdef GETCON

/* A few functions to treat large clusters */

/* Insert entry into large cluster */

void add_to_cluster(int x, int y)
 {
  struct vert_entry *ptr;

  if(x >= width)
   {
    printf("### Error: Coordinate x = %d is not in field ###\n");
    return;
   }

  ptr = vert_space + vert_space_cnt++;
/*
  ptr = (struct vert_entry *)malloc(sizeof(struct vert_entry));
  if(! ptr)
   {
    printf("### Error: Not enough memory to convert large cluster ! ###\n");
    return;
   }
*/

  ptr->y = y;
  ptr->next = large_cluster[x];
  large_cluster[x] = ptr;
 }

/* Free large cluster structure */

void free_large_cluster()
 {
  int i;
  struct vert_entry *ptr;

/*
  for(i=0; i<width; i++)
   if(large_cluster[i])
    {
     for(ptr = large_cluster[i]; ptr; ptr=ptr->next)
       free(ptr);
     large_cluster[i] = 0;
    }
*/
  for(i=0; i<width; i++)
     large_cluster[i] = 0;
  vert_space_cnt = 0;
 }

/* Convert stack into auxiliary structure and then compute
   contribution to connected correlation function */
/* (for large clusters) */

void extract_large_cf_con(long long *cf)
 {
  int i, j;
  int dist;
  INT mask;
  struct vert_entry *ptr1, *ptr2;

  /* Build the structure */

/*
  printf("size = %d\n", stack_cnt);
*/

  for(i=0; i<stack_cnt; i++)
   {
    mask = 1;
    for(j=0; ! (stack[i].mask & mask); j++)
          mask *= 2;
    add_to_cluster((stack[i].column_word << 5) + j , stack[i].row);
   }

/*
  printf(" ... stored\n");
*/

  /* Extract the correlation function from the structure */

  for(i=0; i<width; i++)
   if(large_cluster[i])
    for(ptr1 = large_cluster[i]; ptr1; ptr1 = ptr1->next)
      for(ptr2 = ptr1; ptr2; ptr2 = ptr2->next)
        {
          dist = ptr1->y - ptr2->y;
          if(dist < 0)
            dist = -dist;
          if(dist > height/2)
            dist = height-dist;
          if(dist <= maxcf)
            cf[dist]++;
        }

  /* Free the structure */

  free_large_cluster();
 }


/* Compute contribution from stack to connected correlation function */

void extract_stack_cf_con(long long *cf)
 {
  int i, j;
  int dist;

  for(i=0; i<stack_cnt; i++)
    for(j=i; j<stack_cnt; j++)
     {
      if(stack[i].column_word == stack[j].column_word)
        if(stack[i].mask == stack[j].mask)
         {
           dist = stack[j].row-stack[i].row;
           if(dist < 0)
             dist = -dist;
           if(dist > height/2)
             dist = height-dist;
           if(dist <= maxcf)
             cf[dist]++;
         }
     }
 }

/* Main function for getting connected correlation function */

void get_cf_con(long long *cf, long long *ns, int save_state)
 {
  int i, j;
  int system_size;
  int row;
  int column_word;
  INT mask;
  INT *ptr;

putchar('.');
fflush(stdout);

  system_size = word_width*height;

  if(save_state)
    for(i=0; i<system_size; i++)         /* Save state of system */
       field1[i] = field[i];

  for(column_word=0; column_word < word_width; column_word++)
    for(row=0; row < height; row++)
     {
      ptr = field + word_width*row + column_word;
      for(mask=1<<31; mask; mask = (mask >> 1))
       if(*ptr & mask)
         {
           stack_cnt = 0;
           fill_and_save(ptr, row, column_word, mask);
/*
           if(xdisplay)
             exit_flag |= updateXdisplay();
*/
           ns[stack_cnt-1]++;
           if(stack_cnt < 500)
            {
             extract_stack_cf_con(cf);
            }
           else
            {
             extract_large_cf_con(cf);
            }
         }
     }

  if(save_state)
    for(i=0; i<system_size; i++)         /* Restore state of system */
      field[i] = field1[i];
 }
#endif

/* Output local density fluctuations */

void write_local_dens(long long *ld, char *name, long long norm_fac)
 {
   int i;
   FILE *fpo;

   fpo = fopen(name, "w");
   if(fpo)
    {
     for(i=0; i<=256; i++)
       fprintf(fpo, "%g\t%10lg\n", i/((float) 256), ld[i]/((double) norm_fac));
     fclose(fpo);
    }
 }

/* Output correlation function */
/* 02.05.2000: Better use double than float */
 
void write_cf(long long *cf, char *name, long long norm_fac)
 {
   int i;
   FILE *fpo;
   double rho;
 
   fpo = fopen(name, "w");
   if(fpo)
    {
     rho = cf[0]/((double) norm_fac);
     fprintf(fpo, "# rho = %1.12lg\n", rho);
     for(i=0; i<=maxcf_tp; i++)
       fprintf(fpo, "%d\t%1.12lg\n", i, cf[i]/((double) norm_fac)-rho*rho);
     fclose(fpo);
    }
 }

/* Output connected correlation function */
/* 02.05.2000: Better use double than float */

void write_cf_con(long long *cf, char *name, long long norm_fac)
 {
   int i;
   FILE *fpo;

   fpo = fopen(name, "w");
   if(fpo)
    {
     for(i=0; i<=maxcf; i++)
       fprintf(fpo, "%d\t%1.12lg\n", i, cf[i]/((double) norm_fac));
     fclose(fpo);
    }
 }

/* Output statistics of cluster sizes */
/* 02.05.2000: Better use double than float */
 
void write_ns(long long *ns, int maxsize, char *name, long long norm_fac)
 {
   register long i;
   long last_size;
   FILE *fpo;
 
   last_size = 0;
   for(i=0; i<maxsize; i++)
    if(ns[i])
      last_size = i;
 
   fpo = fopen(name, "w");
   if(fpo)
    {
     for(i=0; (i<maxsize) && (i<=last_size+1); i++)
       fprintf(fpo, "%ld\t%1.12lg\n", i+1, ns[i]/((double) norm_fac));
     fclose(fpo);
    }
 }

/* Dump all necessary information on the status */

void dump_status(long *f_val, int *i, int *maxcf, long *maxsize, INT *field, long long *cf, long long *cf_con, long long *ns, int *local_flag)
 {
  FILE *fp;
  long magic;

  fp = fopen("status", "w");
  if(fp)
   {
    fwrite(f_val, sizeof(long), 1, fp);
    fwrite(i, sizeof(int), 1, fp);
    fwrite(maxcf, sizeof(int), 1, fp);
    fwrite(maxsize, sizeof(long), 1, fp);
    fwrite(&width, sizeof(int), 1, fp);
    fwrite(&height, sizeof(int), 1, fp);
    fwrite(&equilib, sizeof(int), 1, fp);
    fwrite(local_flag, sizeof(int), 1, fp);
    fwrite(field, sizeof(INT), (width*height)>>5, fp);
    fwrite(cf, sizeof(long long), *maxcf+1, fp);
#ifdef GETCON
    fwrite(cf_con, sizeof(long long), *maxcf+1, fp);
    fwrite(ns, sizeof(long long), *maxsize, fp);
#endif
    if(*local_flag)
      fwrite(local_dens, sizeof(long long), 257, fp); 
    magic = MAGICD;
    fwrite(&magic, sizeof(long), 1, fp);
    fclose(fp);
   }
 }

#ifdef GETCON
/* *******************************************************************
   This section contains code for generating cluster quantities from
   a given density fluctuation distribution */

double int_loc_dens[257];

/* Number of configurations to create */

static int configurations=500;

/* A magic number */

#define MAGICG 0x12345678

/* Read the file containing density fluctuations */

void read_fluc(char *name)
 {
   FILE *fpi;
   char temp_str[200];
   register int i;
   int ind;
   double hval, integral;
 
   fpi = fopen(name, "r");
   if(! fpi)
    {
     printf("### Error: Failed to open '%s' for reading. ###\n", name);
     return;
    }
   ind = 0;
   integral = 0;
   while(fgets(temp_str, 200, fpi))
    {
     if(ind < 257)
      {
       i=0;
       while(temp_str[i] && (temp_str[i] <= ' '))
               i++;
       while(temp_str[i] > ' ')
               i++;
       while(temp_str[i] && (temp_str[i] <= ' '))
               i++;
       sscanf(temp_str+i, "%lg", &hval);
       integral += hval;
       int_loc_dens[ind] = integral;
      }
     ind++;
    }
   fclose(fpi);
   if(ind != 257)
     printf("### Error: %d instead of 257 lines in '%s' ###\n", ind, name);
   for(i = 0; i<257; i++)
     int_loc_dens[i] /= integral;
 }

/* Create configuration with a given number of trees */

void create_state(long num_trees)
 {
  int row;
  int column_word;
  int column_bit;
  INT mask;
  INT *ptr;
  long tree_cnt;
  register int i;

/* printf("Create a state with %ld trees\n", num_trees); */

  for(i=0; i < (word_width*height); i++)
     field[i] = 0;

  tree_cnt = 0;
  while(tree_cnt < num_trees)
   {
    row = rand() % height;
    column_word = rand() % word_width;
    column_bit = rand() & 31;
    mask = 1 << column_bit;
    ptr = field + word_width*row + column_word;
    if(! (*ptr & mask))
     {
      *ptr ^= mask;               /* Grow tree */
      tree_cnt++;
     }
   }
 }

/* Dump all necessary information on the status of local density simulations */

void dump_local(int *i, int *maxcf, long *maxsize, long long *cf_con, long long *ns)
 {
  FILE *fp;
  long magic;

  fp = fopen("status_fl", "w");
  if(fp)
   {
    fwrite(i, sizeof(int), 1, fp);
    fwrite(maxcf, sizeof(int), 1, fp);
    fwrite(maxsize, sizeof(long), 1, fp);
    fwrite(&width, sizeof(int), 1, fp);
    fwrite(&height, sizeof(int), 1, fp);
    fwrite(cf_con, sizeof(long long), *maxcf+1, fp);
    fwrite(ns, sizeof(long long), *maxsize, fp);
    fwrite(local_dens, sizeof(long long), 257, fp);
    magic = MAGICG;
    fwrite(&magic, sizeof(long), 1, fp);
    fclose(fp);
   }
 }


/* Main routine for simulating according to a given density distribution */

void do_from_local(long volume, int dump_flag, int eval_flag)
 {
  FILE *fpi;
  double rnum;
  long long *cf_con, *ns;
  int loop_cnt;
  int dens;
  int i;
  long tmp_size;
  long maxsize;
  long magic;

  read_fluc("dens_fluc");

  init_bitcnt();

  if(dump_flag)
   {
    fpi = fopen("status_fl", "r");
    if(! fpi)
     {
      printf("### Error: Failed to open dump file 'status_fl' ! ###\n");
      exit(13);
     }
    else
     {
      printf("Reading from dump file ...\n");
      fread(&loop_cnt, sizeof(int), 1, fpi);
      fread(&maxcf, sizeof(int), 1, fpi);
      fread(&maxsize, sizeof(long), 1, fpi);
      fread(&width, sizeof(int), 1, fpi);
      fread(&height, sizeof(int), 1, fpi);
      word_width = width >> 5;
      volume = width*height;
      loop_cnt++;
     }
   }
  else
   {
    maxsize = volume;           /* Set maximal cluster size */
    if(maxsize > STACKSIZE)
        maxsize = STACKSIZE;
    maxcf = height/2;		/* Set maximal distance for correlation function */
    if(maxcf > MAXCF)
      maxcf = MAXCF;
   }

  field = (INT *) malloc((width*height)/8);
  if(! field)
   {
     printf("### Error: Failed to allocate memory for field !###\n");
     exit(1);
   }

  cf_con = (long long *)malloc(sizeof(long long) * (maxcf + 1));
  if(! cf_con)
   {
     printf("### Error: Not enough memory to store connected correlation function ! ###\n");
     free(field);
     exit(2);
   }

  if(! dump_flag)
    for(i=0; i<=maxcf; i++)
      cf_con[i] = 0;

  ns = (long long *)malloc(sizeof(long long) * maxsize);
  if(! ns)
   {
     printf("### Error: Not enough memory to get statistics of sizes of forest clusters ! ###\n");
     free(cf_con);
     free(field);
     exit(3);
   }

  if(! dump_flag)
    for(i=0; i<maxsize; i++)
      ns[i] = 0;

  large_cluster = (struct vert_entry **)malloc(sizeof(struct vert_entry *) * width);
  if(! large_cluster)
   {
     printf("### Error: Not enough memory to store auxiliary structure for large clusters ! ###\n");
     free(ns);
     free(cf_con);
     free(field);
     exit(3);
   }

  free_large_cluster();

  printf("Width: %d sites\n", width);
  printf("Height: %d sites\n", height);

  if(! dump_flag)
    clear_local_dens();
  else
   {
    fread(cf_con, sizeof(long long), maxcf+1, fpi);
    tmp_size = fread(ns, sizeof(long long), maxsize, fpi);
    if(tmp_size != maxsize)
      printf("### Error: Read %ld instead of %ld elements of n(s). ###\n", tmp_size, maxsize);
    fread(local_dens, sizeof(long long), 257, fpi);
    fread(&magic, sizeof(long), 1, fpi);
    if(magic != MAGICG)
      printf("### Error: Dump file corrupt ! ###\n");
    fclose(fpi);
   }

  if(eval_flag)
   {
    printf("Evaluating contents of dump file 'status_fl' with %d configurations\n", loop_cnt);
    goto eval_it;
   }

  printf("Creating %d configurations according to given density distribution.\n", configurations);

  if(! dump_flag)
   {
     loop_cnt = 0;
   }
  else
   {
    printf("dump already contained %d configurations\n", loop_cnt);
    srand(loop_cnt);
   }

  for(; loop_cnt<configurations; loop_cnt++)
   {
    rnum = (random() % 1000001) / ((double) 1000000);
    for(dens = 0; (dens < 257) && (rnum > int_loc_dens[dens]); dens++)
                            ;
    create_state((volume * dens)/256);
    get_cf_con(cf_con, ns, FALSE);
    local_dens[dens]++;
    if((loop_cnt % 100) == 99)
     {
      dump_local(&loop_cnt, &maxcf, &maxsize, cf_con, ns);
      printf("\n");
     }
   }

eval_it:
  write_local_dens(local_dens, "dens_fluc_fl", loop_cnt);
  write_cf_con(cf_con, "cf_con_plotfile_fl", volume*(long long) loop_cnt);
  write_ns(ns, maxsize, "ns_plotfile_fl", volume*(long long) loop_cnt);

  free(large_cluster);
  free(ns);
  free(cf_con);
  free(field);
 }

#endif

/* Main program */

main(int argc, char *argv[])
 {
  int i, j;
  int argn;
  long volume;
  long maxsize;
  long time_step;
  FILE *fpo, *fpi;
  long long *cf, *cf_con;
  long long *ns;
  int dump_flag;
  int eval_flag;
  int local_flag;
  int maxcf_cut;
  int tmp_val;
  long magic;
#ifdef GETCON
  int from_local;
#endif

  printf("Simulation of the 2D forest-fire model\n");
  printf("Written December 1995 - August 1996 by Andreas Honecker\n");
  printf("(minor modifications April/May 2000)\n");
#ifdef GETCON
  printf("(Version that computes cluster quantities)\n\n");
#else
  printf("(This version does not compute cluster quantities)\n\n");
#endif

  exit_flag = xdisplay = local_flag = eval_flag = dump_flag = FALSE;
  maxcf_cut=MAXCF;
#ifdef GETCON
  from_local = FALSE;
#endif

  for(argn=1; argn<argc; argn++)                                /* Process command line parameters */
   {
     switch(*argv[argn])
      {
        case '-':
          switch(*(argv[argn]+1))
           {
             case 'f':
              sscanf(argv[argn]+2, "%ld", &f_val);
              break; 
             case 'd':
              dump_flag = TRUE;
              break;
             case 'e':
              dump_flag = TRUE;
              eval_flag = TRUE;
              break;
             case 'i':
              sscanf(argv[argn]+2, "%d", &tmp_val);
              iterations = tmp_val * 200;
#ifdef GETCON
              configurations = tmp_val;
#endif
              break;
             case 'x':
              sscanf(argv[argn]+2, "%d", &tmp_val);
              equilib = tmp_val * 200;
              break;
             case 'h':
              sscanf(argv[argn]+2, "%d", &height);
              break;
             case 'w':
              sscanf(argv[argn]+2, "%d", &width);
              break;
             case 'm':
              sscanf(argv[argn]+2, "%d", &maxcf_cut);
              break;
             case 'l':
              local_flag = LOC16X16;
              break;
             case 'L':
              local_flag = LOCNSQ;
              sscanf(argv[argn]+2, "%d", &local_rows);
              break;
             case 's':
              xdisplay = TRUE;
              break;
#ifdef GETCON
             case 'g':
              from_local = TRUE;
              break;
#endif
             default:
              printf("*** Command-Line-Error: Unknown option '%s'\n", argv[argn]);
              goto display_help;
           }
          break;
        default:
         printf("*** Command-Line-Error: Malformed option '%s'\n", argv[argn]);
         goto display_help;
      }
   }

  if(local_flag == LOCNSQ)
   {
    local_columns = 8/local_rows;
    if(local_columns*local_rows != 8)
     {
       printf("*** Error: Only 1, 2, 4 or 8 rows possible\n");
       goto display_help;
     }
   }

  if((width & 31) || (height & 31))
   {
    printf("*** Error: Linear system size must be divisible by 32\n");
    goto display_help;
   }
  if((width <= 0) || (height <= 0))
   {
    printf("*** Error: Linear system size must be positive\n");
    goto display_help;
   }

  word_width = width >> 5;
  volume = width*height;

#ifdef GETCON
  if(from_local)
   {
    do_from_local(volume, dump_flag, eval_flag);
    return(0);
   }
#endif

  if(equilib >= iterations)
   {
    printf("*** Error: Total iterations (%d) should exceed equilibration time (%d)\n", iterations, equilib); 
    goto display_help;
   }
  if(equilib <= 0)
   {
    printf("*** Error: Total iterations (%d) and equilibration time (%d) must be postive !\n", iterations, equilib);
    goto display_help;
   }
  if(iterations > MAX_ITERATIONS)
   {
    printf("*** Error: Total iterations (%d) must be less than %d\n", iterations, MAX_ITERATIONS);
    printf("    In case of need, edit source and change #define MAX_ITERATIONS %d\n", iterations);
    printf("    and recompile\n");
    goto display_help;
   }

  if(dump_flag)
   {
    fpi = fopen("status", "r");
    if(! fpi)
     {
      printf("### Error: Failed to open dump file 'status' ! ###\n");
      exit(13);
     }
    else
     {
      printf("Reading from dump file ...\n");
      fread(&f_val, sizeof(long), 1, fpi);
      fread(&i, sizeof(int), 1, fpi);
      fread(&maxcf, sizeof(int), 1, fpi);
      fread(&maxsize, sizeof(long), 1, fpi);
      fread(&width, sizeof(int), 1, fpi);
      fread(&height, sizeof(int), 1, fpi);
      fread(&equilib, sizeof(int), 1, fpi);
      fread(&local_flag, sizeof(int), 1, fpi);
      word_width = width >> 5;
      volume = width*height;
     }
   }
  else
   {
    maxsize = volume;           /* Set maximal cluster size */
    if(maxsize > STACKSIZE)
         maxsize = STACKSIZE;
   }

  time_step = volume/200;     /* Define how many local iterations constitute a timestep */

  init_bitcnt();

  field = (INT *) malloc((width*height)/8);
  if(! field)
   {
     printf("### Error: Failed to allocate memory for field !###\n");
     exit(1);
   }

  if(! dump_flag)
   {
    maxcf = height/2;
    if(maxcf > MAXCF)
      maxcf = MAXCF;
   }
  maxcf_tp = maxcf;
  if(maxcf_tp > maxcf_cut)
    maxcf_tp = maxcf_cut;

  cf = (long long *)malloc(sizeof(long long) * (maxcf + 1));
  if(! cf)
   {
     printf("### Error: Not enough memory to store correlation function ! ###\n");
     free(field);
     exit(2);
   }
#ifdef GETCON
  cf_con = (long long *)malloc(sizeof(long long) * (maxcf + 1));
  if(! cf_con)
   {
     printf("### Error: Not enough memory to store connected correlation function ! ###\n");
     free(cf);
     free(field);
     exit(2);
   }
#endif

  if(! dump_flag)
   {
    for(i=0; i<=maxcf; i++)
     {
      cf[i] = 0;
#ifdef GETCON
      cf_con[i] = 0;
#endif
     }
   }

#ifdef GETCON
  field1 = (INT *) malloc((width*height)/8);
  if(! field1)
   {
     printf("### Error: Failed to allocate memory for copy of field !###\n");
     free(cf_con);
     free(cf);
     free(field);
     exit(3);
   }

  ns = (long long *)malloc(sizeof(long long) * maxsize);
  if(! ns)
   {
     printf("### Error: Not enough memory to get statistics of sizes of forest clusters ! ###\n");
     free(field1);
     free(cf_con);
     free(cf);
     free(field);
     exit(3);
   }

  if(! dump_flag)
   for(i=0; i<maxsize; i++)
     ns[i] = 0;

  large_cluster = (struct vert_entry **)malloc(sizeof(struct vert_entry *) * width);
  if(! large_cluster)
   {
     printf("### Error: Not enough memory to store auxiliary structure for large clusters ! ###\n");
     free(ns);
     free(field1);
     free(cf_con);
     free(cf);
     free(field);
     exit(3);
   }

/*
  for(i=0; i<width; i++)
   {
    large_cluster[i] = 0;
   }
*/
  free_large_cluster();
#endif

/*
  for(i=0; i < (width*height)/32; i++)
   {
     field[i] = i;
     if(! (i & 15))
        field[i] = 0x7fffffff;
     if((i & 15) == 1)
        field[i] = 257;
     if((i & 15) == 2)
        field[i] = 257+65536;
   }
*/

/* Code for testing fill */

/*
  for(i=0; i < (width*height)/32; i++)
   {
     field[i] = 1<<31;
     if(! ((i / word_width) & 0x1f))
       field[i] = 0xffff0fff;
   }
  for(i=0; i<word_width; i++)
     field[i] = 0xffffffff;
  if(xdisplay)
   {
  initXdisplay(argc, argv, "2D forest-fire model", field, width, height);
  sleep(3);
  exit_flag |= updateXdisplay();
   }
  sleep(10);

  fill(field, 0, 0, 1<<31);

  if(xdisplay)
    exit_flag |= updateXdisplay();
  sleep(10);
  exit(0);
*/

  printf("f_val = %ld\n", f_val);
  printf("(f = %10g)\n", f_val/((float) 1000000));
  printf("Equilibration time: %d (= %d global time_steps)\n", equilib, equilib/200);
  printf("Total iteration time: %d (= %d global time_steps)\n", iterations, iterations/200);
  printf("Width: %d sites\n", width);
  printf("Height: %d sites\n", height);
  if(local_flag == LOC16X16)
     printf("Extracting local density fluctuations on 16x16 plaquettes\n");
  if(local_flag == LOCNSQ)
     printf("Extracting local density fluctuations on %dx%d plaquettes\n", 32*local_columns, local_rows);

  if(! dump_flag)
   {
    init_field();
    if(local_flag)
      clear_local_dens();
   }
  else
   {
    fread(field, sizeof(INT), (width*height)>>5, fpi);
    fread(cf, sizeof(long long), maxcf+1, fpi);
#ifdef GETCON
    fread(cf_con, sizeof(long long), maxcf+1, fpi);
    fread(ns, sizeof(long long), maxsize, fpi);
#endif
    if(local_flag)
      fread(local_dens, sizeof(long long), 257, fpi);
    fread(&magic, sizeof(long), 1, fpi);
    if(magic != MAGICD)
      printf("### Error: Dump file corrupt ! ###\n"); 
    fclose(fpi);
   }

  if(xdisplay)
   {
    initXdisplay(argc, argv, "2D forest-fire model", field, width, height);
    printf("\nDisplaying state of the system in window\n");
    printf("It is strongly recommended to use the console of the computer for this purpose\n");
    printf("\nPress `P' inside the state window to create a binary file `snapshot.lj' that \n");
    printf("\tcan be printed on a HP LaserJet (or other printer understanding HPGL).\n");
    printf("Press `Q' inside the state window to suppress displaying of the state of the system.\n");
    printf("Press `S' to re-enable displaying the current state of the system.\n");
    printf("Press `X' to terminate simulation.\n");
    exit_flag |= updateXdisplay();
   }

  if(eval_flag)
   {
    printf("Evaluating contents of dump file 'status' with i = %d\n", i);
    goto eval_it;
   }

  if(! dump_flag)
   {
    fpo = fopen("conc", "w");
    i = 0;
   }
  else
   {
    fpo = fopen("conc", "a");
    printf("starting after dump at i = %d\n", i);	/* Output of i added 02.05.2000 */
   }

  for(; (i<iterations) && (! exit_flag); i++)
   {
    if((! (i % 200)) && i && (! dump_flag))
     {
       dump_status(&f_val, &i, &maxcf, &maxsize, field, cf, cf_con, ns, &local_flag);
       if(fpo)
        {
         for(j=i-200; j<i; j++)
              fprintf(fpo, "%ld\t%10g\n", j+1, conc[j]/((float) volume));
         fflush(fpo);
        }
     }
    dump_flag = FALSE;

    conc[i] = count_trees();
    if(xdisplay)
      displayXdens(conc[i], volume);

    if(i >= equilib)
     if(! ((i-equilib) % 200))
       {
         if(local_flag == LOC16X16)
           get_local_dens16x16();
         if(local_flag == LOCNSQ)
           get_local_dens_cr();

         get_cf(field, cf);
#ifdef GETCON
         get_cf_con(cf_con, ns, TRUE);
#endif
       }
    for(j=0; j<time_step; j++)
      iteration_step();
    if(xdisplay)
      exit_flag |= updateXdisplay();
   }

  if(fpo)
    fclose(fpo);

/*
  fpo = fopen("conc", "w");
 
  if(fpo)
   {
    for(i=0; i<iterations; i++)
      fprintf(fpo, "%ld\t%10g\n", i+1, conc[i]/((float) volume));
    fclose(fpo);
   }
*/

eval_it:
  if(i > equilib)
   {
    write_cf(cf, "cf_plotfile", volume*(long long)(i-equilib)/200);
    if(local_flag)
      write_local_dens(local_dens, "dens_fluc", volume*(long long)(i-equilib)/(200*256));
#ifdef GETCON
    write_cf_con(cf_con, "cf_con_plotfile", volume*(long long)(i-equilib)/200);
    write_ns(ns, maxsize, "ns_plotfile", volume*(long long)(i-equilib)/200);
#endif
   }

  if(xdisplay)
    exitXdisplay();
#ifdef GETCON
  free(large_cluster);
  free(ns);
  free(field1);
  free(cf_con);
#endif
  free(cf);
  free(field);
  return(0);		/* Normal Termination */

display_help:
  printf("\nUSE: %s [options]\n", argv[0]);
  printf("Valid options are:\n");
  printf("-f#\tSet f/p to #/1000000 (default: %ld)\n", f_val);
  printf("-d\tStart after the point specified in the status files\n");
  printf("-e\tEvaluate information stored in the status files\n");
  printf("-i#\tSet number of total iterations to # global time steps (default: %d)\n", iterations/200);
  printf("-x#\tSet number of initial iterations to discard for equilibration\n");
  printf("\tto # global time steps (default: %d)\n", equilib/200);
  printf("-h#\tSet system height to # sites (default: %d)\n", height);
  printf("-w#\tSet system width to # sites (default: %d)\n", width);
  printf("-m#\tSet maximal distance for evaluation of two-point function to #\n");
  printf("\t(Used to save CPU time)\n");
  printf("-l\tExtract local density fluctuations on 16x16 plaquettes\n");
  printf("-L#\tExtract local density fluctuations on 256/# times # plaquettes\n");
  printf("-s\tShow state of the system on X display\n");
#ifdef GETCON
  printf("-g\tGenerate cluster distributions from local density fluctuations in 'dens_fluc'\n");
  printf("\t(options -f -x -m -l -s are meaningless if -g is specified)\n");
  printf("\t(-i# together with -g specifies number of configurations to be generated [default: %d])\n",
         configurations);
#endif
 }
