#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "xdisplay.h"
#include "printlj.h"


/* Values of truth */

#ifndef TRUE
  #define TRUE -1
#endif
#ifndef FALSE
  #define FALSE 0
#endif

/* Width of a boundary (around the fields) */

#define BOUND_WIDTH 16

/* Height of status part */

#define STATUS_HEIGHT 64

/* Characters x and y offset */

#define CXOFS 4
#define CYOFS 5

/* Size of the field */

static int field_width=512;
static int field_height=512;

/* Flag if displaying is actually to be used */

static int display_flag = TRUE;

char disp_string1[] = "Displaying of the state of the system is disabled.";
char disp_string2[] = "Press `S' to re-enable it.";

/* Variables needed for X11 access */

static XImage *fieldimage=NULL;
static Display *mydisplay;
static Window mywindow;
static GC mygc, inv_gc, text_gc;
static XFontStruct *my_font;

/* The density line */

int *x_dens;
int xdens_cnt = 0;
int time_cnt = 0;
XPoint *x_points;

/* Close display */

void exitXdisplay()
 {
  XFreeGC(mydisplay, mygc);
  XDestroyWindow(mydisplay, mywindow);
  XCloseDisplay(mydisplay);
  XFree(fieldimage);
  free(x_points);
  free(x_dens);
 }

/* Show the complete density */

void show_x_dens()
 {
  int i;

/*
  XFillRectangle(mydisplay, mywindow, inv_gc,
                 BOUND_WIDTH+1, field_height+2*BOUND_WIDTH+1, field_width-2, STATUS_HEIGHT-BOUND_WIDTH-2);
*/
  for(i=0; i<xdens_cnt; i++)
   {
    x_points[i].x = BOUND_WIDTH+i;
    x_points[i].y = field_height+BOUND_WIDTH+STATUS_HEIGHT-x_dens[i];
   }
  XDrawPoints(mydisplay, mywindow, mygc, x_points, xdens_cnt, CoordModeOrigin);
 }

/* Scroll the complete density */

void scroll_x_dens()
 {
  int xa;
  int ya;
  int i;

/*
  for(i=1; i<xdens_cnt-1; i++)
   {
    x_points[i-1].x = BOUND_WIDTH+i;
    x_points[i-1].y = field_height+BOUND_WIDTH+STATUS_HEIGHT-x_dens[i];
   }
  XDrawPoints(mydisplay, mywindow, inv_gc, x_points, xdens_cnt-2, CoordModeOrigin);

  for(i=0; i<xdens_cnt-1; i++)
   {
    x_dens[i] = x_dens[i+1];
    x_points[i].x = BOUND_WIDTH+i;
    x_points[i].y = field_height+BOUND_WIDTH+STATUS_HEIGHT-x_dens[i];
   }
  XDrawPoints(mydisplay, mywindow, mygc, x_points, xdens_cnt-1, CoordModeOrigin);
*/
  for(i=0; i<xdens_cnt-1; i++)
   if(x_dens[i] != x_dens[i+1])
    {
     xa = BOUND_WIDTH+i;
     ya = field_height+BOUND_WIDTH+STATUS_HEIGHT-x_dens[i];
     if(i)
       XDrawPoint(mydisplay,mywindow, inv_gc, xa, ya);
     x_dens[i] = x_dens[i+1];
     ya = field_height+BOUND_WIDTH+STATUS_HEIGHT-x_dens[i];
     XDrawPoint(mydisplay,mywindow, mygc, xa, ya);
    }
 }

/* add a density */

void displayXdens(long trees, long volume)
 {
  int value;
  int xa;
  int ya;
  int i;

  value = (trees*(STATUS_HEIGHT-BOUND_WIDTH))/volume;
  if(xdens_cnt < field_width)
   {
    x_dens[xdens_cnt] = value;
    xa = BOUND_WIDTH+xdens_cnt++;
    ya = field_height+BOUND_WIDTH+STATUS_HEIGHT-value;
    XDrawPoint(mydisplay,mywindow, mygc, xa, ya);
   }
  else
   {			/* We have to scroll */
    scroll_x_dens();
    x_dens[field_width-1] = value;
    xa = BOUND_WIDTH+xdens_cnt;
    ya = field_height+BOUND_WIDTH+STATUS_HEIGHT-value;
    XDrawPoint(mydisplay,mywindow, mygc, xa, ya);
   }
/*
  if(! (time_cnt & 0xf))
    XFlush(mydisplay);
   time_cnt++;
*/
 }

/* Clear field display */

void quiet_field_display()
 {
  int slen;

  XFillRectangle(mydisplay, mywindow, mygc,
             BOUND_WIDTH, BOUND_WIDTH, field_width, field_height);
  slen = XTextWidth(my_font, disp_string1, strlen(disp_string1));
  XDrawImageString(mydisplay, mywindow, text_gc,
                   BOUND_WIDTH+(field_width-slen)/2, field_height/2, disp_string1, strlen(disp_string1));
  slen = XTextWidth(my_font, disp_string2, strlen(disp_string2));
  XDrawImageString(mydisplay, mywindow, text_gc,
                   BOUND_WIDTH+(field_width-slen)/2, 2*BOUND_WIDTH+field_height/2, disp_string2, strlen(disp_string2));
 }

/* Update display */

int updateXdisplay()
 {
  XEvent myevent;
  KeySym mykey;
  char text[10];
  int done;
  int i;
  int xa, ya;
  int exit_flag;

  exit_flag = FALSE;

  if(display_flag)
    XPutImage(mydisplay,mywindow,mygc,fieldimage,0,0,
              BOUND_WIDTH, BOUND_WIDTH, field_width,field_height);

  while(XEventsQueued(mydisplay, QueuedAlready))
   {
    XNextEvent(mydisplay, &myevent);
    switch(myevent.type)
     {
      case Expose:
       if(! myevent.xexpose.count)
         if(display_flag)
           XPutImage(mydisplay,mywindow,mygc,fieldimage,0,0,
                     BOUND_WIDTH, BOUND_WIDTH, field_width,field_height);
         else
           quiet_field_display();
         xa = BOUND_WIDTH;
         ya = field_height+2*BOUND_WIDTH;
         XDrawLine(mydisplay,mywindow,mygc,
                   xa, ya, xa, ya-BOUND_WIDTH+STATUS_HEIGHT);
         XDrawLine(mydisplay,mywindow,mygc,
                   xa, ya, xa+field_width, ya);
         XDrawImageString(mydisplay, mywindow, mygc,
                          xa-2*CXOFS, ya+CYOFS, "1", 1);
         XDrawImageString(mydisplay, mywindow, mygc,
                          xa-2*CXOFS, ya-BOUND_WIDTH+STATUS_HEIGHT+CYOFS, "0", 1);
         xa = BOUND_WIDTH+field_width;
         XDrawLine(mydisplay,mywindow,mygc,
                   xa, ya, xa, ya-BOUND_WIDTH+STATUS_HEIGHT);
         XDrawImageString(mydisplay, mywindow, mygc,
                          xa-CXOFS, ya-BOUND_WIDTH+STATUS_HEIGHT+2*CYOFS, "t", 1);
         xa = BOUND_WIDTH;
         ya = field_height+BOUND_WIDTH+STATUS_HEIGHT;
         XDrawLine(mydisplay,mywindow,mygc,
                   xa, ya, xa+field_width, ya);
         show_x_dens();
       break;
      case MappingNotify:
       XRefreshKeyboardMapping(&myevent);
       break;
      case ButtonPress:
/*
       if(myevent.xbutton.button == 1)              */ /* Left Button */ /*
         XDrawImageString(myevent.xbutton.display, myevent.xbutton.window, mygc,
                          myevent.xbutton.x, myevent.xbutton.y, hi, strlen(hi));
       if(myevent.xbutton.button == 3)             */ /* Right Button */ /*
         XDrawPoint(myevent.xbutton.display, myevent.xbutton.window, mygc,
                    myevent.xbutton.x, myevent.xbutton.y);
*/
       break;
      case KeyPress:
       i = XLookupString(&myevent, text, 10, &mykey, 0);
       if(i == 1)
         switch(text[0] | 0x20)
           {
            case 'p':
             print_lj("snapshot.lj", maximize_dpi(field_width, field_height),
                      (INT *) fieldimage->data, field_width, field_height);
             break;
            case 'q':
             display_flag = FALSE;
             quiet_field_display();
             break;
            case 's':
             display_flag = TRUE;
             XPutImage(mydisplay,mywindow,mygc,fieldimage,0,0,
                       BOUND_WIDTH, BOUND_WIDTH, field_width,field_height);
             break;
            case 'x':
             exit_flag = TRUE;
             break;
           }
       break;
     }
   }
  return(exit_flag);
 }

/* init display */

void initXdisplay(int argc, char *argv[], char *title, INT *field, int width, int height)
 {
  XSizeHints myhint;
  int myscreen;
  unsigned long myforeground, mybackground;
  int i;
  int pid;
  INT tempv;

  field_width = width;
  field_height = height;

  x_dens = (int *)malloc(sizeof(int)*width);
  x_points = (XPoint *)malloc(sizeof(XPoint)*width);

  mydisplay = XOpenDisplay("");
  if(! mydisplay)
   {
     printf("### Error: Failed to open display !###\n");
     exit(1);
   }

  myscreen = DefaultScreen(mydisplay);
  mybackground = WhitePixel(mydisplay, myscreen);
  myforeground = BlackPixel(mydisplay, myscreen);

  myhint.x = 50;
  myhint.y = 40;
  myhint.width = field_width+2*BOUND_WIDTH;
  myhint.height = field_height+2*BOUND_WIDTH+STATUS_HEIGHT;
  myhint.flags = PPosition | PSize;

  mywindow = XCreateSimpleWindow(mydisplay, DefaultRootWindow(mydisplay),
                                 myhint.x, myhint.y, myhint.width, myhint.height,
                                 5, myforeground, mybackground);
  XSetStandardProperties(mydisplay, mywindow, title, title, None, argv, argc,
                         &myhint);
  mygc=XCreateGC(mydisplay, mywindow, 0, 0);
  XSetBackground(mydisplay, mygc, mybackground);
  XSetForeground(mydisplay, mygc, myforeground);
  inv_gc=XCreateGC(mydisplay, mywindow, 0, 0);
  XSetBackground(mydisplay, inv_gc, myforeground);
  XSetForeground(mydisplay, inv_gc, mybackground);

  text_gc=XCreateGC(mydisplay, mywindow, 0, 0);
  XSetBackground(mydisplay, text_gc, myforeground);
  XSetForeground(mydisplay, text_gc, mybackground);
  my_font = XLoadQueryFont(mydisplay, "9x15");
  XSetFont(mydisplay, text_gc, my_font->fid);

  XSelectInput(mydisplay, mywindow, ButtonPressMask | KeyPressMask | ExposureMask);
  XMapRaised(mydisplay, mywindow);

  fieldimage =XCreateImage(mydisplay,(Visual *) &mywindow,1,XYBitmap,0,
    (char*) field, field_width, field_height, 32, 0);


  tempv = *field;
  /* Test for byte order on client */
  *field=1;
  if( * (char*) field)
   {
    fieldimage->byte_order=LSBFirst;
    /* printf("byte_order=LSBFirst\n"); */
   }
  else
   {
    fieldimage->byte_order=MSBFirst;
    /* printf("byte_order=MSBFirst\n"); */
   }
#ifdef TESTBITORD
  /* Find out in what order bits are put on screen */
  *field=0xff;
  if(XGetPixel(fieldimage,0,0))
   {
    fieldimage->bitmap_bit_order=MSBFirst;
    printf("bitmap_bit_order=MSBFirst\n");
   }
  else
   {				/* bits are backward */
    fieldimage->bitmap_bit_order=LSBFirst;
    printf("bitmap_bit_order=LSBFirst\n"); 
   }
#else
  /* All CPUs tested have the most significant bit first */
  fieldimage->bitmap_bit_order=MSBFirst;
#endif
  *field = tempv;

  updateXdisplay();
 }
